import io
import os
import sys
import logging
from contextlib import redirect_stdout, redirect_stderr
from typing import List
from enum import Enum
from jinja2 import Environment, PackageLoader, select_autoescape

from termcolor import colored
import yaml
from PIL import Image

from staq.function import Function, FunctionArgument

from html2image import Html2Image


rich_color_map = {
    'grey': 'gray',
    'red': 'red',
    'green': 'green',
    'yellow': 'yellow',
    'blue': 'blue',
    'magenta': 'magenta',
    'cyan': 'cyan',
    'white': 'white',
    'dark_grey': 'grey53',
    'dark_red': 'maroon',
    'dark_green': 'dark green',
    'dark_yellow': 'olive',
    'dark_blue': 'navy',
    'dark_magenta': 'purple',
    'dark_cyan': 'teal',
    'dark_white': 'silver'
}

class StackWord():
    def __init__(self, cell, value = None):
        self.cell = cell
        self.value = value
        self.note = None
        self.noteColor = 'red'
        self.address = None

    def __str__(self):
        if self.value:
            return str(self.value)
        else:
            return " "
    

    def setNote (self, note, color = 'red'):
        self.note = note
        self.noteColor = color
        
    def isHead(self):
        return self.address == self.cell.frame.stack.pointer
    
    def shouldShow(self):
        return self.value or self.note or self.isHead() or (self.address == self.cell.address) or self.address in self.cell.frame.stack.argMarks 

    def toAnsi(self, width = 80, leftMargin = 20, color = "blue", showAddress = True):
        
        out = ""

        lmargin = " "*leftMargin

            

        lAddress = colored(f"{lmargin}")


        if showAddress and self.address:
            lAddress = colored(f"{hex(self.address).rjust(leftMargin)}", color="dark_grey")
            pass
        
        if self.isHead():
            strlen =  leftMargin - (len(str(hex(self.address))) + 3)
            pointer = colored(">>>", color="yellow") 
            strAddress = colored(hex(self.address), color="dark_grey")
            lAddress = " "*strlen + pointer + strAddress 
        
        out += lAddress + colored(f"│{str(self).ljust(width-2)}│", color=color)

        if self.address in self.cell.frame.stack.argMarks:
            out += colored(f"{self.cell.frame.stack.argMarks[self.address]}", "black","on_green")

        if self.note:
            out += colored(f" <-- {self.note}", color=self.noteColor)
        
        out += "\n"

        return out



class StackCell():
    def __init__(self, label = "", words=[], note=None, frame=None):
        self.address = None
        self.frame = None
        self.words: List[StackWord] = []
        self.size = 1
        self.label = label
        self.color = None

        if words:
            self.setWords(words)
            self.size = len(words)



    def setSize(self, size):
        self.size = size 
        while(len(self.words) < size):
            self.words.append(StackWord(self))

    def setWords(self,words, offset = 0):



        #if not a list, make it a list
        if not isinstance(words, list):
            words = [words]

        self.setSize(len(words))



        for word in words:

            if word != "" and word != " " and offset < self.size:

                if isinstance(word, StackWord):
                    self.words[offset] = word
                else:
                    self.words[offset].value = word
                offset += 1


    def setNote (self, note):
        self.words[0].note = note

    def applyAddresses(self, baseAddress):

        self.address = baseAddress

        for word in self.words:
            word.address = baseAddress
            baseAddress += (self.frame.stack.bits // 8)

        return baseAddress
    
    @classmethod
    def fromObj(cls, obj):

        ret = StackCell()

        if not isinstance(obj, dict):
            ret.setWords(obj)
        
        else:

            key = list(obj.keys())[0]
            value = obj[key]

            ret.label = key

            if not isinstance(value, dict):
                ret.setWords(value)

            else:

                if 'words' in value:
                    ret.setWords(value['words'])
                else:
                    ret.words = []

                if 'color' in value:
                    ret.color = value['color']

                if 'note' in value:
                    ret.note = value['note']

                if 'size' in value:
                    ret.setSize(value['size'])



        return ret
    
    def toAnsi(self, width = 80, leftMargin = 20, color = "blue", showAddress = True, sof = False):

        out = ""

        lmargin = " "*leftMargin

        lAddress = colored(f"{lmargin}")

        if showAddress and self.address:
            lAddress = colored(f"{hex(self.address).rjust(leftMargin)}", color="dark_grey")
            pass

        if self.color:
            color = self.color

        top = f"├{'─'*(width-2)}┤"

        header = self.label

        if self.size > 1:
            header = f"{self.label} [{self.size * (self.frame.stack.bits // 8)}]"

        if header and header != "":
            
            top = f"├{header.center(width-2,'─')}┤"

            topParts = top.split(header)

            top = colored(topParts[0], color=color) + colored(header, color="dark_grey") + colored(topParts[1], color=color)

            if sof:
                top = f"┌{header.center(width-2,'─')}┐"
        
        else:
            if sof:
                top = f"┌{'─'*(width-2)}┐"
        

        out += colored(f"{lmargin}{top}\n", color=color)


        showIndexes = []
        emptyStreak = 0

        for i in range(self.size):

            if self.words[i].shouldShow() :
                showIndexes.append(i)
                emptyStreak = 0
            else:
                emptyStreak += 1

            if emptyStreak == 1:
                showIndexes.append(None)

            # if emptyStreak < 2:
            #     showIndexes.append(i)

        for i in showIndexes:

            if i is None:
                out+= colored(f"{lmargin}:{' '*(width-2)}:\n", color=color)
            else: 
                word = self.words[i]

                out+= word.toAnsi(width, leftMargin, color, showAddress)




        return out




class StackFrame():
    def __init__(self, name=None, function :Function = None):
        self.name = name
        self.function = function
        self.cells: List[StackCell] = []
        self.color = None
        self.stack = None
        self.basePointer = 0
        

    def length(self):
        return len(self.cells)

    def pop(self):


        if len(self.cells) == 0:
            return False
        
        ret = "???"
        cell = self.cells[-1]

        if cell.size == 1:
            if len(cell.words) > 0:
                ret = cell.words.pop(0)
            self.cells.pop()
        else :
            if(len(cell.words) > 0):
                ret = cell.words.pop(0)
            cell.size -= 1
        
        return ret
        
    
    def push(self, cell):
        
        cell.frame = self
        self.cells.append(cell)

    def applyAddresses(self, baseAddress = 0x07f00000):

        for cell in self.cells:
            baseAddress = cell.applyAddresses(baseAddress)
 
        
        return baseAddress
            

    @classmethod
    def fromObj(cls, obj, order = 'normal'):
            
        ret = cls()

        if 'function' in obj:
            ret.function = obj['function']
            ret.color = "blue"
        
        if 'stack' in obj:
            cells = []
            if order == 'normal':
                cells = reversed(obj['stack'])
            else:
                cells = obj['stack']

            for cell in cells:
                ret.push(StackCell.fromObj(cell))

        if 'color' in obj:
            ret.color = obj['color']

        return ret
    
    def toAnsi(self, width = 80, leftMargin = 20, color= "blue", showAddress = False):

        #           ╭─main──────────────────────────────╮
        #  0x070000 │         i = 0x014                 │
        #           ├─────────────buffer────────────────┤
        #  0x070000 :        ...                        :
        #           ├─────────────arg1──────────────────┤
        #  0x070000 │         32                        | 
        #           ├──────────────arg1─────────────────┤
        #  0x070000 │         14                        │
        #           ╰──────────────arg2─────────────────╯

        out = ""

        lmargin = " "*leftMargin
        fill = "─"*(width - 2)

        if not self.function:
            color = "dark_grey"

        if self.color:
            color = self.color


        if self.function:
            fill = "─"*(width - (len(self.function.name) + 3))
            out += colored(f"{lmargin}╭─{self.function.name}{fill}╮\n", color=color)
        # else: 
        #     out += colored(f"{lmargin}┌{fill}┐\n",color=color)
        
        for i,cell in enumerate( reversed(self.cells)):
           
           sof = False
           if i == 0 and not self.function:
                sof = True
           out += cell.toAnsi(width, leftMargin, color, sof=sof,showAddress=showAddress)

        if self.function:
            fill = "─"*(width - 2)
            out += colored(f"{lmargin}╰{'─'*(width - 2)}╯\n", color=color)
        else:
            out += colored(f"{lmargin}└{fill}┘\n",color=color)

        return out
        


class Stack():
    def __init__(self, baseAddress = 0xffff):
        self.frames: List[StackFrame] = []
        self.currentFrame = None
        self.bits = 32
        self.baseAddress = baseAddress
        self.endian = 'little'
        self.pointer = self.baseAddress
        self.argMarks = {}

    def setBaseAddress(self, baseAddress):

        offset = baseAddress - self.baseAddress
        self.pointer += offset
        self.baseAddress = baseAddress

        self.applyAddresses()

    def nextAddress(self):

        offset = self.bits // 8
        return self.pointer + offset


    def movePointer(self, wordCount):

        byteCount = wordCount * (self.bits // 8)
        self.pointer += byteCount

    def resolveAddress(self, address):

        if isinstance(address, str):
            if address.startswith("&"):
                var = address[1:]
                offset = 0
                if '+' in var:
                    parts = var.split('+')
                    var = parts[0]
                    offset = int(parts[1])
                
                if '-' in var:
                    parts = var.split('-')
                    var = parts[0]
                    offset = -1 * int(parts[1])


                for frame in reversed(self.frames):
                    for cell in reversed(frame.cells):
                        if cell.label == var:
                            return cell.address + offset


            elif address.startswith("0x"):
                return int(address, 16)
            else:
                return int(address)
        else:
            return address
        
    def getCell(self, address):
            
        for frame in self.frames:
            for cell in frame.cells:

                start = cell.address
                end = cell.address+ (cell.size * (self.bits//8))

                if (address >= start) and (address < end):
                    return cell

            
        return None
        
    def getWord(self, address):
            
            cell = self.getCell(address)
            if cell:
                offset = int((address - cell.address) // (self.bits // 8))

                return cell.words[offset]
            else:
                return None

    def write(self, values, address = None):

        if not isinstance(values, list):
            values = [values]

        if not address:
            address = self.pointer
        else:
            address = self.resolveAddress(address)
        
        for value in values:
            

            word = self.getWord(address)
            if word:
                word.value = value

                address += (self.bits // 8)

    def setNote (self, note, address = None, color = 'red'):

        if not address:
            address = self.pointer
        else:
            address = self.resolveAddress(address)

        word = self.getWord(address)
        if word:
            word.setNote(note,color)
    
    def popFrame(self):
        
        if self.currentFrame:

            count = 0

            for cell in self.currentFrame.cells:
                count += cell.size 

            self.movePointer(count)
        
            self.frames.pop()

            if len(self.frames ) > 0:
                self.currentFrame = self.frames[-1]
            else:
                self.currentFrame = None

    def pushFrame(self, frame):
        frame.stack = self
        frame.basePointer = self.pointer -1
        self.frames.append(frame)
        self.currentFrame = self.frames[-1]

    def clear(self):
        self.frames = []
        self.currentFrame = None
        self.pointer = self.baseAddress

    def popLocals(self):

        if self.currentFrame:

            while self.currentFrame.length() > 1:
                self.currentFrame.pop()

    def pop(self, count = 1):

        ret = None
        while count > 0:
            if self.currentFrame:

                ret = self.currentFrame.pop()
                if ret is not None:
                    self.movePointer(1)
                    count -= 1
                if self.currentFrame.length() == 0:
                    self.popFrame()
            else:

                return ret


            
        
        return ret

    
    def push(self, cell: StackCell):

        if not self.currentFrame:
            self.pushFrame(StackFrame())

        self.currentFrame.push(cell)
        self.movePointer(-1 * cell.size)

        self.currentFrame.cells[-1].applyAddresses(self.pointer)


        
    def call(self, function , args):

        if isinstance(function, Function):
            self.frames.append(StackFrame(function.name))
        else:
            self.frames.append(StackFrame(function))

    def applyAddresses(self, baseAddress = None):

        if not baseAddress:
            baseAddress = self.baseAddress

        self.nextAddr = baseAddress
        for frame in self.frames:
            self.nextAddr = frame.applyAddresses(self.nextAddr)

        return self.nextAddr
    
    def loadYaml(self, obj):
        

        if isinstance(obj, str):
            with open(obj) as f:
                obj = yaml.safe_load(f)
        
        if 'stackBase' in obj:
            self.baseAddress = obj['stackBase']
        
        if 'order' in obj:
            order = obj['order']

        if 'stack' in obj:

            nodes = []
            if order == 'normal':
                nodes = reversed(obj['stack'])
            else:
                nodes = obj['stack']

            for node in nodes:
                
                if isinstance(node, dict) and 'function' in node:
                    #Create a new frame 
                    frame = StackFrame.fromObj(node, order= order)
                    self.pushFrame(frame)
                    self.currentFrame = None
                else:
                        cell = StackCell.fromObj(node)
                        self.push(cell)
        
        self.applyAddresses()

    def getLeftMargin(self, showAddress = True):

        margin = 3 # room for pointer >>> 

        if showAddress:
            
            h = hex(self.baseAddress) #this is the longest it will be 
            margin += len(h) + 2 #room for 0x

        return margin

    
    def toAnsi(self, width = 50, showAddress = False, leftMargin = None):
     
        # Another style 
        #           ╭─vuln──────────────────────────────╮
        #  0x070000 │                                   │
        #           ├───────────────────────────────────┤
        #  0x070000 │                                   | 
        #           ├───────────────────────────────────┤
        #  0x070000 │   <main+054>                      │ <- return to main
        #           ╰───────────────────────────────────╯
        #           ╭─main──────────────────────────────╮
        #           ├───────────────i───────────────────┤
        #  0x070000 │ 0x014                             │
        #           ├─────────────buffer────────────────┤
        #           :                                   : 
        #  0x070000 :  ...                              : 
        #           ├───────────────arg1────────────────┤
        #  0x070000 │ 17                                | 
        #           ├───────────────arg2────────────────┤
        #  0x070000 │ 38                                │
        #           ╰───────────────────────────────────╯
        #           ┌───────────────────────────────────┐
        #  0x070000 │     . . .                         │
        #           └───────────────────────────────────┘


        if leftMargin == None:
                leftMargin = self.getLeftMargin(showAddress)

        out = ""
        for i,frame in enumerate(reversed(self.frames)):
            
            color = "blue"
            if frame == self.currentFrame:
                color = "green"

            out += frame.toAnsi(width,color=color, showAddress=showAddress, leftMargin=leftMargin)

        #Add end of stack

        if showAddress:
            out+= colored(f"{hex(self.baseAddress).rjust(leftMargin)}", color="dark_grey")
        else:
            out += colored(f'{" "*(leftMargin)}', color="dark_grey")
        out += colored(f' {"End Of Stack".center((width-2), "─")}\n', color="dark_grey")

        return out


    

    
    def toHtml(self, showAddress = True, full = True):
        env = Environment(
            loader=PackageLoader('staq', 'templates'),
            autoescape=select_autoescape(['html', 'xml'])
        )
        template = env.get_template('stack.html.j2')
        if full:
            template = env.get_template('stack-full.html.j2')
        html_output = template.render(stack=self, showAddress = showAddress)
        return html_output

    
    def print(self, showAddress= True, leftMargin = 20):
        
        ascii = self.toAnsi(showAddress=showAddress, leftMargin=leftMargin)
        print(ascii)
        
    
    def generatePng(self, filePath = 'out.png', width = None):

        html = self.toHtml()

        ansi = self.toAnsi( showAddress=False)
        width = width 


        lines = ansi.count('\n')

        path, name = os.path.split(filePath)


        #This will fail if chrome is not installed
        hti = Html2Image()



        # Create a string buffer to capture stdout and stderr
        f = io.StringIO()

        # Redirect stdout and stderr
        with redirect_stdout(f), redirect_stderr(f):
            hti.output_path = path
            hti.screenshot(html_str=html, save_as=name)

        #crop the image
        with Image.open(filePath) as img:
            img = img.convert("RGBA")
            
            #crop the image to remove the extra whitespace
            bbox = img.getbbox()

            #add some padding to all side 
            padding = 10
            bbox = (bbox[0] - padding, bbox[1] - padding, bbox[2] + padding, bbox[3] + padding)


            img = img.crop(bbox)

            if width:
                currWidth = img.width
                currHeight = img.height

                #assume pixels
                newWidth = width
                newHeight = int((newWidth/currWidth) * currHeight)

                if isinstance(width, str):
                    if width.endswith("%"):
                        newWidth = int((int(width[:-1])/100) * currWidth)
                        newHeight = int((int(width[:-1])/100) * currHeight)
                    if width.endswith("px"):
                        newWidth = int(width[:-2])
                        newHeight = int((newWidth/currWidth) * currHeight)

                img = img.resize((newWidth, newHeight))

                    


            img.save(filePath)


        





from staq.stack import Stack, StackFrame, StackCell 
from staq.function import Function, FunctionArgument
from staq.conventions import CallingConvention, Register
from typing import List

import re

class CdeclConvention(CallingConvention):
    def __init__(self, session = None, stack = None):
        super().__init__('cdecl', session=session, stack=stack)
        self.endian = 'little'


        self.registers = {
            'eax': Register('eax', 4, 'Return Value'),
            'ecx': Register('ecx', 4, 'Arg 1'),
            'edx': Register('edx', 4, 'Arg 2'),
            'ebx': Register('ebx', 4, 'Arg 3'),
            'esp': Register('esp', 4, 'Stack Pointer'),
            'ebp': Register('ebp', 4, 'Base Pointer'),
            'esi': Register('esi', 4, 'Source Index'),
            'edi': Register('edi', 4, 'Destination Index'),
            'eip': Register('eip', 4, 'Instruction Pointer')
        }

        self.setReg('esp', hex(self.stack.pointer))
        self.setReg('eip', '???')
        self.setReg('ebp', hex(self.stack.baseAddress))

    def clear(self):

        for key in self.registers:
            self.registers[key].value = None
        
        self.setReg('esp', hex(self.stack.pointer))
        self.setReg('eip', '???')
        self.setReg('ebp', hex(self.stack.baseAddress))

    # def leave(self):

    #     if self.registers['ebp'].value and self.stack:
            
    #         esp = self.stack.pointer
    #         ebp = self.getReg('ebp')

    #         #pop everything until the prev ebp
    #         while esp < ebp:
    #             self.stack.pop()
    #             esp = self.stack.pointer

    #         prev_ebp = self.stack.pop()
    #         self.setReg("ebp",prev_ebp.value)
    #         self.setReg("esp",esp)




    def setInstructionPointer(self, val):

        self.setReg("eip", val)
    
    


        



from staq.conventions import CdeclConvention
from staq.stack import Stack, StackFrame, StackCell
from staq.function import Function, FunctionArgument, parseCFile
from staq.helper import splitWithEscapes
import re
import yaml
import json
import argparse

import importlib.resources as pkg_resources
import builtins



sizeDict = {
    "void": 0,
    "char": 1,
    "unsigned char": 1,
    "signed char": 1,
    "short": 2,
    "unsigned short": 2,
    "int": 4,
    "unsigned int": 4,
    "long": 4,
    "unsigned long": 4,
    "long long": 8,
    "unsigned long long": 8,
    "float": 4,
    "double": 8,
    "long double": 16,  # This can vary, but 16 bytes is a common size
    "int8": 1,
    "uint8": 1,
    "int16": 2,
    "uint16": 2,
    "int32": 4,
    "uint32": 4,
    "int64": 8,
    "uint64": 8,
    "va_list": 4,
}

def nullPrint(*args, **kwargs):
    pass

builtins.print = nullPrint

class StackSession():
    def __init__(self):
        self.stack = Stack()
        self.conv = CdeclConvention(session=self, stack = self.stack)
        self.functions = {}
        self.parser = argparse.ArgumentParser(description='Stack Visualizer', add_help=False)
        self.subparsers = None
        self.init_args()
        self.loadLibcFuntions()
        self.commands = []
        self.update = False
        self.status = ""
        self.cmdIdx = 0

    def addHistory(self,cmd):

        if self.cmdIdx < len(self.commands):
            self.commands = self.commands[:self.cmdIdx]

        self.commands.append(cmd)
        self.cmdIdx+=1

    def stepForward(self):

        if self.cmdIdx < len(self.commands):
            self.parseCommand(self.commands[self.cmdIdx], addHistory=False)
            self.cmdIdx +=1

    def stepBack(self):

        if self.cmdIdx > 0:

            self.cmdIdx -= 1
            self.stack.clear()
            
            for i in range(0,self.cmdIdx):
                self.parseCommand(self.commands[i], addHistory=False)

    def print(self,text):
        print(text)

    def refreshOutput(self):

        ansi = self.stack.toAnsi( showAddress=True)
        print(ansi)

    def loadYaml(self, obj):


        if isinstance(obj, str):
            with open(obj) as f:
                obj = yaml.safe_load(f)


        order = 'normal'

        if 'stackBase' in obj:
            self.stack.baseAddress = obj['stackBase']
        
        if 'order' in obj:
            order = obj['order']

        if 'functions' in obj:
            for func in obj['functions']:
                self.addFunction(Function.fromString(func))
    

        if 'stack' in obj:

            nodes = []
            if order == 'normal':
                nodes = reversed(obj['stack'])
            else:
                nodes = obj['stack']

            for node in nodes:
                
               if isinstance(node, dict) and 'function' in node:
                   #Create a new frame 
                   frame = StackFrame.fromObj(node, order= order)
                   self.stack.pushFrame(frame)
                   self.stack.currentFrame = None
               else:
                    cell = StackCell.fromObj(node)
                    self.stack.push(cell)
        
        self.stack.applyAddresses()


    
    def getCurrentFunction(self):

        if self.stack.currentFrame:

            if self.stack.currentFrame.function:
                return self.stack.currentFrame.function
            
        
        return None

    def addFunction(self, function):
        self.functions[function.name] = function


    def loadLibcFuntions(self):
        
        data_path = pkg_resources.files('staq.data').joinpath('libc.c')

        with open(data_path) as f:
            functions = parseCFile(f.read())
            for f in functions:
                self.addFunction(f)

        # for f in libc_functions:
        #     newFunc = Function.fromString(f)

        #     self.addFunction(newFunc)

    def parseFrame(self,words):
        
        line = " ".join(words[1:])
        self.stack.pushFrame(StackFrame(line))

    def frameCmd(self, args):
        frame = StackFrame(args.name)
        if args.color:
            frame.color = args.color
        self.stack.pushFrame(frame)

    def popCmd(self, args):

    
        if args.count and isinstance(args.count,str):
            if args.count.lower() == 'frame':
                self.stack.popFrame()
            elif args.count.lower().replace("$","") in self.conv.registers:

                word = self.stack.pop()
                self.conv.setReg(args.count.lower().replace("$",""), word.value )
            else:
                count = int(args.count)
                self.stack.pop(count)
        else:
            self.stack.pop()

    def retCmd(self, args):

        self.conv.ret()

    def jmpCmd(self, args):

        self.conv.jmp(args.address)

    def pushCmd(self, args):
        cell = StackCell()

        if args.value:

            line = " ".join(args.value)

            parts = line.split(":")

            if len(parts) > 1:
                cell.label = parts[0]

                words = parts[1].strip().split(",")
                words = [x.strip() for x in words]

                cell.setWords(words)
            else:
                words = parts[0].split(",")
                words = [x.strip() for x in words]
                cell.setWords(parts[0].split(","))


        if args.size:
            cell.setSize(args.size)
        
        if args.label:
            cell.label = args.label

        if args.address:
            cell.address = args.address

        if args.note:

            note = " ".join(args.note)

            cell.setNote(note)



        self.stack.push(cell)


    def parseFunction(self, line):

        function = Function.fromString(line)
        self.addFunction(function)

   
    def callCmd(self, args):
            

            line = " ".join(args.function)
    
            self.conv.call(line)

    def leaveCmd(self, args):
        self.conv.leave()

    def functionCmd(self,args):
        line = " ".join(args.function)
        function = Function.fromString(line)
        self.addFunction(function)
    
    def tryParseLocalVar(self, line):
        
        parts = line.split("=")

        
        revar = re.compile(r"(\w[\w\s\*]*\w)\s+(\w+)(\[(\d*)\])?")

        ptr = False

        decl = parts[0]
        if "*" in decl:
            ptr = True
            decl = decl.replace("*","")

        match = revar.match(decl)
        val = []
        if len(parts) > 1:
            parts[1] = parts[1].split("--")[0]
            val = parts[1].replace(";","")
            val = yaml.safe_load(val)

        if match:
            varType = match.group(1).strip()

            varName = match.group(2)

            arraySize = match.group(4) if match.group(4) else None

            cell = StackCell(varName)

            if "_t" in varType:
                varType = varType.replace("_t","")

            if varType not in sizeDict:
                return None

            
            if ptr:
                sizeBytes = 4
            else:
                sizeBytes = sizeDict.get(varType, 4)

            if arraySize:
                sizeBytes = sizeBytes * int(arraySize)
            
            size = int((sizeBytes + 3) / 4)

            cell.setWords(val)
            cell.setSize(size)
            cell.label = varName


            line = line.replace(match.group(0), 'local')


            return cell

        else:
            return None
            
    def localCmd(self, cell, args):

        if args.note:
            cell.setNote(" ".join(args.note))

            if args.note_color:
                cell.words[0].noteColor = args.note_color

        if args.color:
            cell.color = args.color

        self.stack.push(cell)

    def saveCmd(self, args):

        filename = args.filename




        if filename.endswith('.html'):
            with open(filename, "w") as f:
                f.write(self.stack.toHtml())

        elif filename.endswith('.png'):

            try:
                self.stack.generatePng(filename)
            except Exception as e:
                self.print("Failed to generate image. This is likely due to not having 'chrome' installed which is required for html2image. Please install chrome and try again.")

        else:
            with open(filename, "w") as f:
                for cmd in self.commands:
                    if not cmd.startswith("save"):
                        if not cmd.endswith("\n"):
                            cmd += "\n"
                        f.write(cmd)


                
        


    def loadCmd(self, args):
        filename = args.filename
        self.update = False

        if filename.endswith('.c'):

            with open(filename) as f:
                functions = parseCFile(f.read())

            self.status = f"Loaded {len(functions)} functions from {filename}"

            for func in functions:
                self.addFunction(func)

    def writeCmd(self, args):
        """
        
        write data1,data,2,data3 --address &cell+4
        write &cell+16: data1,data2,data3
        write &cell+16: [data1,data2,data3] 
        
        """
        address  = None
        line = " ".join(args.values)
        valLine = line
        
        parts = splitWithEscapes(line,":")

        if len(parts) > 1:
            address = parts[0]
            valLine = parts[1]


        if not valLine.startswith("[") and not valLine.endswith("]"):
            valLine = f"[{valLine}]"

        values = yaml.safe_load(valLine) #use yaml to parse the list so we get the correct types/escapes


        if args.address:
            address = args.address

        self.stack.write(values, address)


    def noteCmd(self, args):

        address  = args.address
        line = " ".join(args.note)
        note = line
        
        parts = splitWithEscapes(line,":")

        if len(parts) > 1:
            address = parts[0]
            note = parts[1]



        self.stack.setNote(note, address, args.color)

    def run(self, callLine = None, limit = 20):

        called = True

        functionCall = None
        
        if callLine:
            callLine = callLine.strip()
            if callLine == "":
                callLine = None

        if callLine:
            functionCall = callLine
        elif self.stack.currentFrame and self.stack.currentFrame.function:
            function = self.stack.currentFrame.function

            if len(function.calls) > 0:
                functionCall  = function.calls[0]

        if not functionCall:
            functionCall = "main()"

        self.conv.call(functionCall)
        limit -= 1

        while  called and (limit > 0):
            
            if self.stack.currentFrame and self.stack.currentFrame.function: 
                function = self.stack.currentFrame.function

                if len(function.calls) > 0:
                    call = function.calls[0]
                    self.conv.call(call)
                    called = True

            limit -= 1

    def runCmd(self, args):

        call = None
        if args.function:
            call = " ".join(args.function)


        self.run(call, limit = args.limit)

    def clearCmd(self, args):
        self.stack.clear()

    def parseCommand(self, line, addHistory = True):

        plainCmd = line.split(" ")[0]


        if addHistory and plainCmd not in ['save','s', 'load','l']:
            self.addHistory(line)
        self.update = True
        self.status = ""

        commands = splitWithEscapes(line,";")

        for cmd in commands:

            local = self.tryParseLocalVar(cmd)

            if local:
                cmd = f"local {cmd}"
                args = self.parser.parse_args(cmd.split(" "))
                self.localCmd(local, args)
                self.refreshOutput()

            else:

                try:
                    args = self.parser.parse_args(cmd.split(" "))
                    #print(args)

                    if args.help:
                        self.showHelp(args.command)
                    else:
                        args.func(args)
                        self.refreshOutput()
                except:
                    pass
                    words = cmd.split(" ")
                    self.showHelp(words[0])


    def showHelp(self, command):
            
            if command in self.subparsers.choices:
                self.print(self.subparsers.choices[command].format_help() + "\n\n [Enter] to continue...")
            else:
                self.print(self.parser.format_help() + "\n\n [Enter] to continue...")

            #self.print("[Enterr] to continue...")



    # Initialize the argument parser
    def init_args(self):

        self.parser.add_argument('--help', '-h', action='store_true', help='show this help message and exit')

        subparsers = self.parser.add_subparsers(dest='command', help='sub-command help')

        self.parser.error = nullPrint
        #local command 
        local_parser = subparsers.add_parser('local', help='Declaure a local variable using c syntax')
        local_parser.add_argument('local', nargs='*', help='Local variable declaration')
        local_parser.add_argument('--note','-n', nargs='*', type=str, help='Note for the cell')
        local_parser.add_argument('--color', '-c', type=str, help="Color of the cell")
        local_parser.add_argument('--note-color', '-nc', type=str, help="Color of the note")
        local_parser.set_defaults(func=self.localCmd)

        # pop command
        pop_parser = subparsers.add_parser('pop', help='pop words from the stack')
        pop_parser.add_argument('count', type=str, nargs='?', help='Number of elements to pop', default=1)
        pop_parser.set_defaults(func=self.popCmd)

        #ret command
        ret_parser = subparsers.add_parser('ret', help='Return (pop, jmp to return address)')
        ret_parser.set_defaults(func=self.retCmd)

        #jmp command
        jmp_parser = subparsers.add_parser('jmp', help='Jump to an address')
        jmp_parser.add_argument('address',  help='Address to jump to')
        jmp_parser.set_defaults(func=self.jmpCmd)
        
        #push command
        push_parser = subparsers.add_parser('push', help='push words to the stack')
        push_parser.add_argument('value', nargs='*', help='Value to push', default='')
        push_parser.add_argument('--size', '-s', type=int, help='Type of value')
        push_parser.add_argument('--label','-l', help='Label for the value')
        push_parser.add_argument('--address','-a', help='Address for the value')
        push_parser.add_argument('--note','-n', nargs='*', help='Note for the value')

        push_parser.set_defaults(func=self.pushCmd)

        #frame command
        frame_parser = subparsers.add_parser('frame', help='Start a new Frame')
        frame_parser.add_argument('name', nargs="?",  help='Name of the frame [omit to close the current frame]')
        frame_parser.add_argument('--color', '-c', help='Color of the frame')
        frame_parser.add_argument('--address', '-a', help='address of the frame ')
        frame_parser.add_argument('--size', '-s', type=int, help='size of the frame')
        frame_parser.set_defaults(func=self.frameCmd)

        #call parser
        call_parser = subparsers.add_parser('call', help='Call a function')
        call_parser.add_argument('function', nargs="*", help='Function to call')
        call_parser.set_defaults(func=self.callCmd)

        #function 
        function_parser = subparsers.add_parser('function', help="Define a function using c syntax")
        function_parser.add_argument('function', nargs="*", help='Function signature')
        function_parser.set_defaults(func=self.functionCmd)

        #save 
        save_parser = subparsers.add_parser('save', help='Save the current stack')
        save_parser.add_argument('filename', nargs="?", type=str, help='Filename to save', default='.stackinit')
        save_parser.set_defaults(func=self.saveCmd)

        #load
        load_parser = subparsers.add_parser('load', help='load stack file')
        load_parser.add_argument('filename', type=str, help='Filename to load')
        load_parser.set_defaults(func=self.loadCmd)

        #run 
        run_parser = subparsers.add_parser('run', aliases=['r'], help='Run the a function calls the `call` within a function')
        run_parser.add_argument('function', nargs="*", help='Function call to start, if not provided will use the first call in the current function, or main if no function is active')
        run_parser.add_argument('--limit', '-l', type=int, help='Limit of calls to make', default=20)
        run_parser.set_defaults(func=self.runCmd)

        #write 
        write_parser = subparsers.add_parser('write', aliases=['w'], help='Write data to stack')
        write_parser.add_argument('values', nargs="*", help='Values to write')
        write_parser.add_argument('--address', '-a', type=str, help='Address to write to can be a reference to a cell &<cellName>+4')
        write_parser.add_argument('--count', '-c', type=int, help='Repeated writes', default=1)
        write_parser.add_argument('--file', '-f', type=str, help='binary file to write')
        write_parser.set_defaults(func=self.writeCmd)


        #note
        note_parser = subparsers.add_parser('note', help='Add a note to the stack')
        note_parser.add_argument('note', nargs="*", help='Note to add')
        note_parser.add_argument('--color', '-c', type=str, help='Color of the note', default='red')
        note_parser.add_argument('--address', '-a', type=str, help='Address of the note', default=None)
        note_parser.set_defaults(func=self.noteCmd)

        #leave 
        leave_parser = subparsers.add_parser('leave', help='Pops words from stack until the previous base pointer')
        leave_parser.set_defaults(func=self.leaveCmd)


        clear_parser = subparsers.add_parser('clear', help='Clears the stack')
        clear_parser.set_defaults(func=self.clearCmd)

        for sub in subparsers.choices:
            subparsers.choices[sub].print_help =  nullPrint
            subparsers.choices[sub].error = nullPrint


        self.subparsers = subparsers



from docutils import nodes
from docutils.parsers.rst import Directive, directives
from sphinx.util.docutils import SphinxDirective

import os


from staq.stack import Stack
from staq.stackSession import StackSession
import yaml
import time 


class StackDirective(SphinxDirective):
    has_content = True

    optional_arguments = 1
    option_spec = {
        'file': directives.path,
        'baseaddress': directives.positive_int,
        'showaddresses': directives.unchanged,
        'width': directives.unchanged,
    }


    def run(self):
        base_address = self.options.get('baseaddress', 0xffff)
        show_addresses = self.options.get('showaddresses', 'true')
        width = self.options.get('width', '100%')

        if show_addresses.lower() == 'true':
            show_addresses = True
        else:
            show_addresses = False


        # Get the Sphinx builder
        builder = self.state.document.settings.env.app.builder.name
        session = StackSession()
        session.stack.setBaseAddress(base_address)

        contents = ""

        if self.arguments:
            filePath = self.arguments[0]

            if not os.path.exists(filePath):
                error = nodes.error(None, nodes.paragraph(text=f"File not found: {filePath}"))
                return [error]

            with open(filePath, "r") as fh:
                contents = fh.read()

        else:
            contents= '\n'.join(self.content)


        contentHash = str(abs(hash(contents)))
        output_dir = self.state.document.settings.env.app.outdir

        image_name = f"stack_{contentHash[:8]}.png"
        image_path = os.path.join(output_dir, image_name)
        #log_path = os.path.join(output_dir, "stack.log")

        # with open(log_path, "w") as fh:
        #     fh.write(f"content hash: {contentHash}\n")
        #     fh.write(f"image path: {image_path}\n")
        #     fh.write(f"builder: {builder}\n")
        #     fh.write(f"output dir: {output_dir}\n")
        #     fh.write(f"base address: {base_address}\n")
        #     fh.write(f"show address: {show_addresses}\n")
        #     fh.write(f"content: {contents}\n")

        for line in contents.split('\n'):
            session.parseCommand(line)

        if builder == 'html':
            html_content = session.stack.toHtml(showAddress= show_addresses, full=False)
            raw_html_node = nodes.raw('',html_content, format='html')
            return [raw_html_node]
        else:
    

            timeout = 10

            try:
                session.stack.generatePng(image_path, width=width)

                while not os.path.exists(image_path) and timeout > 0:
                    timeout -= 1
                    time.sleep(1)

                if not os.path.exists(image_path):
                    raise FileNotFoundError(f"Image not generated: {image_path}\ntry again")
            except Exception as e:
                print(f"Error: {e}")
                error = nodes.error(None, nodes.paragraph(text=f"Error: {e}"))

                with open("error.log", "w") as fh:
                    fh.write(str(e))

                return [error]

            image_node = nodes.image(uri= f"/{image_path}")
            # image_node['width'] = width

            return [image_node]
        

    



def generate_stack_image(stack, base_address, show_addresses, env, contentHash):
    

    image_path = os.path.join(env.app.outdir, f"stack_{contentHash}.png")
    stack.generatePng(image_path)

    
    return image_path

def setup(app):
    app.add_directive('stack', StackDirective)

    return {'parallel_read_safe': True}
